@@ -34,9 +34,21 @@ end
 
 module Setup : sig
 
+  (* State extensions see {!module:State} *)
+  type state_descriptor
+
+  (* Quotation extensions see {!module:Quotations} *)
+  type quotations_descriptor
+
+  (* HOAS encoding extensions see {!module:RawData} *)
+  type hoas_descriptor
+
   (* Built-in predicates, see {!module:BuiltIn} *)
   type builtins
 
+  (* Operations avaiable via calc (infix is), see {!module:Calc} *)
+  type calc_descriptor
+
   (* Compilation flags, see {!module:Compile} *)
   type flags
   
@@ -60,6 +72,10 @@ module Setup : sig
         [file_resolver]. *)
   val init :
     ?flags:flags ->
+    ?state:state_descriptor ->
+    ?quotations:quotations_descriptor ->
+    ?hoas:hoas_descriptor ->
+    ?calc:calc_descriptor ->
     builtins:builtins list ->
     ?file_resolver:(?cwd:string -> unit:string -> unit -> string) ->
     ?legacy_parser:bool ->
@@ -145,6 +161,7 @@ module Data : sig
     state : state;
     output : 'a;
     pp_ctx : pretty_printer_context;
+    relocate_assignment_to_runtime : target:state -> depth:int -> string -> (term, string) Stdlib.Result.t (* uvars are turned into discard *)
   }
 
   (* Hypothetical context *)
@@ -544,6 +561,8 @@ module BuiltInPredicate : sig
   type 'a oarg = Keep | Discard
   type 'a ioarg = private Data of 'a | NoData
 
+  type once
+
   type ('function_type, 'inernal_outtype_in, 'internal_hyps, 'internal_constraints) ffi =
     (* Arguemnts that are translated independently of the program context *)
     | In    : 't Conversion.t * doc * ('i, 'o,'h,'c) ffi -> ('t -> 'i,'o,'h,'c) ffi
@@ -562,6 +581,7 @@ module BuiltInPredicate : sig
       context readback function *)
     | Read : ('h,'c) ContextualConversion.ctx_readback * doc -> (depth:int -> 'h -> 'c -> Data.state -> 'o, 'o,'h,'c) ffi
     | Full : ('h,'c) ContextualConversion.ctx_readback * doc -> (depth:int -> 'h -> 'c -> Data.state -> Data.state * 'o * Conversion.extra_goals, 'o,'h,'c) ffi
+    | FullHO : ('h,'c) ContextualConversion.ctx_readback * doc -> (once:once -> depth:int -> 'h -> 'c -> Data.state -> Data.state * 'o * Conversion.extra_goals, 'o,'h,'c) ffi
     | VariadicIn    : ('h,'c) ContextualConversion.ctx_readback * ('t,'h,'c) ContextualConversion.t * doc -> ('t list -> depth:int -> 'h -> 'c -> Data.state -> Data.state * 'o, 'o,'h,'c) ffi
     | VariadicOut   : ('h,'c) ContextualConversion.ctx_readback * ('t,'h,'c) ContextualConversion.t * doc -> ('t oarg list -> depth:int -> 'h -> 'c -> Data.state -> Data.state * ('o * 't option list option), 'o,'h,'c) ffi
     | VariadicInOut : ('h,'c) ContextualConversion.ctx_readback * ('t ioarg,'h,'c) ContextualConversion.t * doc -> ('t ioarg list -> depth:int -> 'h -> 'c -> Data.state -> Data.state * ('o * 't option list option), 'o,'h,'c) ffi
@@ -625,6 +645,51 @@ module BuiltInPredicate : sig
 
   end
 
+  (** Adaptors for standard HO functions *)
+  module HOAdaptors : sig
+
+    type 'a pred1
+    type ('a,'b) pred2
+    type ('a,'b,'c) pred3
+
+    val pred1 : 'a Conversion.t -> 'a pred1 Conversion.t
+    val pred2 : 'a Conversion.t -> 'b Conversion.t -> ('a,'b) pred2 Conversion.t
+    val pred3 : 'a Conversion.t -> 'b Conversion.t -> 'c Conversion.t -> ('a,'b,'c) pred3 Conversion.t
+
+    val filter1 :
+      once:once -> depth:int ->
+      filter:(('a -> bool) -> 's -> 's) ->
+      'a pred1 ->
+      's ->
+      Data.state ->
+      Data.state * 's * Conversion.extra_goals
+
+    val filter2 :
+      once:once -> depth:int ->
+      filter:(('a -> 'b -> bool) -> 's -> 's) ->
+      ('a,'b) pred2 ->
+      's ->
+      Data.state ->
+      Data.state * 's * Conversion.extra_goals
+
+    val map1 :
+      once:once -> depth:int ->
+      map:(('a -> 'c) -> 's -> 's) ->
+      ('a,'c) pred2 ->
+      's ->
+      Data.state ->
+      Data.state * 's * Conversion.extra_goals
+
+    val map2 :
+      once:once -> depth:int ->
+      map:(('a -> 'b -> 'c) -> 's -> 's) ->
+      ('a,'b,'c) pred3 ->
+      's ->
+      Data.state ->
+      Data.state * 's * Conversion.extra_goals
+
+  end
+
 end
 
 (** Setup.init takes a list of declarations of data types and predicates,
@@ -667,8 +732,8 @@ module BuiltIn : sig
   (** Prints in LP syntax the "external" declarations.
     * The file builtin.elpi is generated by calling this API on the
     * declaration list from elpi_builtin.ml *)
-  val document_fmt : Format.formatter -> Setup.builtins -> unit
-  val document_file : ?header:string -> Setup.builtins -> unit
+  val document_fmt : Format.formatter -> ?calc:Setup.calc_descriptor -> Setup.builtins -> unit
+  val document_file : ?header:string  -> ?calc:Setup.calc_descriptor -> Setup.builtins -> unit
 
 end
 
@@ -728,6 +793,8 @@ end
    need to use. *)
 module State : sig
 
+  val new_state_descriptor : unit -> Setup.state_descriptor
+
   (** 'a MUST be purely functional, i.e. backtracking is implemented by using
    * an old binding for 'a.
    * This limitation can be lifted if there is user request. *)
@@ -740,7 +807,18 @@ module State : sig
     (* run just before the goal is compiled (but after the program is) *)
     start:('a -> 'a) ->
       'a component
+  [@@deprecated "Use [declare_component] instead"]
 
+  val declare_component :
+    ?descriptor:Setup.state_descriptor ->
+    name:string ->
+    pp:(Format.formatter -> 'a -> unit) ->
+    init:(unit -> 'a) ->
+    (* run just before the goal is compiled (but after the program is) *)
+    start:('a -> 'a) ->
+    unit ->
+      'a component
+  
   type t = Data.state
 
   val get : 'a component -> t -> 'a
@@ -913,6 +991,32 @@ module RawOpaqueData : sig
   val to_loc : t -> Ast.Loc.t
   val of_loc : Ast.Loc.t -> Data.term
 
+end
+
+module Calc : sig
+
+  type operation_declaration = {
+    symbol : string;
+    infix : bool; (* used for the doc *)
+    args : string list list; (* multiple types for the same symbol *)
+    code : Data.term list -> Data.term;
+  }
+
+  (** Registering an operation *)
+  val register : descriptor:Setup.calc_descriptor -> operation_declaration -> unit
+
+  (** An empty descriptor for registering operations *)
+  val new_calc_descriptor : unit -> Setup.calc_descriptor
+
+  (** Standard operations *)
+  val default_calc : operation_declaration list
+
+  (** The [calc] and [is] declarations *)
+  val calc : BuiltIn.declaration list
+
+  (** for use in other builtins *)
+  val eval : depth:int -> State.t -> Data.term -> Data.term
+
 end
 
   (** This module exposes the low level representation of terms.
@@ -1022,8 +1126,11 @@ module RawData : sig
      Since extension to the data type extra_goal are global to all elpi
      instances, this post-processing function is also global *)
   val set_extra_goals_postprocessing :
+    ?descriptor:Setup.hoas_descriptor ->
     (Conversion.extra_goals -> State.t -> State.t * Conversion.extra_goals) -> unit
 
+  val new_hoas_descriptor : unit -> Setup.hoas_descriptor
+
 end
 
 (** This module lets one generate a query by providing a RawData.term directly *)
@@ -1039,10 +1146,16 @@ module RawQuery : sig
   (* Args are parameters of the query (e.g. capital letters). *)
   val is_Arg : State.t -> Data.term -> bool
 
+  (* with the possibility to update the state in which the query will run *)
+  val compile_ast :
+    Compile.program -> Ast.query -> (State.t -> State.t) -> unit Compile.query
+
+  (* generate the query term and initial state by hand *)
   val compile :
     Compile.program -> (depth:int -> State.t -> State.t * (Ast.Loc.t * Data.term) * Conversion.extra_goals) ->
       unit Compile.query
 
+  
 end
 
 module Quotation : sig
@@ -1051,10 +1164,10 @@ module Quotation : sig
     depth:int -> State.t -> Ast.Loc.t -> string -> State.t * Data.term
 
   (** The default quotation [{{code}}] *)
-  val set_default_quotation : quotation -> unit
+  val set_default_quotation : ?descriptor:Setup.quotations_descriptor -> quotation -> unit
 
   (** Named quotation [{{name:code}}] *)
-  val register_named_quotation : name:string -> quotation -> unit
+  val register_named_quotation : ?descriptor:Setup.quotations_descriptor -> name:string -> quotation -> unit
 
   (** The anti-quotation to lambda Prolog *)
   val lp : quotation
@@ -1072,12 +1185,14 @@ module Quotation : sig
    * needs something that looks like a string but with a custom compilation
    * (e.g. CD.string like but with a case insensitive comparison) *)
 
-  val declare_backtick : name:string ->
+  val declare_backtick : ?descriptor:Setup.quotations_descriptor -> name:string ->
     (State.t -> string -> State.t * Data.term) -> unit
 
-  val declare_singlequote : name:string ->
+  val declare_singlequote : ?descriptor:Setup.quotations_descriptor -> name:string ->
     (State.t -> string -> State.t * Data.term) -> unit
 
+  val new_quotations_descriptor : unit -> Setup.quotations_descriptor
+
 end
 
 module Utils : sig
@@ -1106,7 +1221,7 @@ module Utils : sig
 
   (** Hackish, in particular the output should be a compiled program *)
   val clause_of_term :
-    ?name:string -> ?graft:([`After | `Before] * string) ->
+    ?name:string -> ?graft:([`After | `Before | `Replace] * string) ->
     depth:int -> Ast.Loc.t -> Data.term -> Ast.program
 
   (** Lifting/restriction/beta (LOW LEVEL, don't use) *)
