@@ -20,7 +20,7 @@ exception CompileError of Loc.t option * string
 type builtins = string * Data.BuiltInPredicate.declaration list
 
 type header
-val header_of_ast : flags:flags -> parser:(module Parse.Parser) -> builtins list -> Ast.Program.t ->  header
+val header_of_ast : flags:flags -> parser:(module Parse.Parser) -> State.descriptor -> QuotationHooks.descriptor -> HoasHooks.descriptor -> CalcHooks.descriptor -> builtins list -> Ast.Program.t ->  header
 
 type program
 val program_of_ast : flags:flags -> header:header -> Ast.Program.t -> program
@@ -31,7 +31,7 @@ val assemble_units : flags:flags -> header:header -> compilation_unit list -> pr
 val append_units : flags:flags -> base:program -> compilation_unit list -> program
 
 type 'a query
-val query_of_ast : program -> Ast.Goal.t -> unit query
+val query_of_ast : program -> Ast.Goal.t -> (State.t -> State.t) -> unit query
 val query_of_term :
   program -> (depth:int -> State.t -> State.t * (Loc.t * term) * Conversion.extra_goals) -> unit query
 val query_of_data :
@@ -40,17 +40,14 @@ val query_of_data :
 val optimize_query : 'a query -> 'a executable
 
 val term_of_ast : depth:int -> State.t -> string -> State.t * term
+val relocate_closed_term : from:State.t -> to_:State.t -> term -> (term, string) Stdlib.Result.t
 
 val pp_program : (pp_ctx:pp_ctx -> depth:int -> Format.formatter -> term -> unit) -> Format.formatter -> 'a query -> unit
 val pp_goal : (pp_ctx:pp_ctx -> depth:int -> Format.formatter -> term -> unit) -> Format.formatter -> 'a query -> unit
 
 val lookup_query_predicate : program -> string -> program * Data.constant
 
-type quotation = depth:int -> State.t -> Loc.t -> string -> State.t * term
-val set_default_quotation : quotation -> unit
-val register_named_quotation : name:string -> quotation -> unit
-
-val lp : quotation
+val lp : QuotationHooks.quotation
 
 val is_Arg : State.t -> term -> bool
 val get_Args : State.t -> term StrMap.t
@@ -70,9 +67,6 @@ val static_check :
 
 module CustomFunctorCompilation : sig
 
-  val declare_singlequote_compilation : string -> (State.t -> F.t -> State.t * term) -> unit
-  val declare_backtick_compilation : string -> (State.t -> F.t -> State.t * term) -> unit
-
   val compile_singlequote : State.t -> F.t -> State.t * term
   val compile_backtick : State.t -> F.t -> State.t * term
 
