@@ -53,6 +53,7 @@ module Constants : sig
   module Set : Set.S with type elt = constant
   val pp : Format.formatter -> t -> unit
   val show : t -> string
+  val compare : t -> t -> int
 end = struct
 
 module Self = struct
@@ -115,6 +116,25 @@ let uvar_isnt_a_blocker { uid_private } = uid_private > 0 [@@inline];;
 let uvar_set_blocker r   = r.uid_private <- -(uvar_id r) [@@inline];;
 let uvar_unset_blocker r = r.uid_private <-  (uvar_id r) [@@inline];;
 
+type arg_mode = Util.arg_mode = Input | Output [@@deriving show]
+
+type clause = {
+    depth : int;
+    args : term list;
+    hyps : term list;
+    vars : int;
+    mode : mode;        (* CACHE to avoid allocation in get_clauses *)
+    loc : Loc.t option; (* debug *)
+}
+and 
+mode = arg_mode list
+[@@deriving show]
+
+let to_mode = function true -> Input | false -> Output
+
+
+module DT = Discrimination_tree
+
 type stuck_goal = {
   mutable blockers : blockers;
   kind : unification_def stuck_goal_kind;
@@ -138,25 +158,22 @@ and second_lvl_idx =
 | TwoLevelIndex of {
     mode : mode;
     argno : int;
-    all_clauses : clause list;         (* when the query is flexible *)
-    flex_arg_clauses : clause list;       (* when the query is rigid but arg_id ha nothing *)
-    arg_idx : clause list Ptmap.t;   (* when the query is rigid (includes in each binding flex_arg_clauses) *)
+    all_clauses : clause list;        (* when the query is flexible *)
+    flex_arg_clauses : clause list;   (* when the query is rigid but arg_id ha nothing *)
+    arg_idx : clause list Ptmap.t;    (* when the query is rigid (includes in each binding flex_arg_clauses) *)
   }
 | BitHash of {
     mode : mode;
     args : int list;
-    time : int; (* time is used to recover the total order *)
+    time : int;                             (* time is used to recover the total order *)
     args_idx : (clause * int) list Ptmap.t; (* clause, insertion time *)
   }
-and clause = {
-    depth : int;
-    args : term list;
-    hyps : term list;
-    vars : int;
-    mode : mode; (* CACHE to avoid allocation in get_clauses *)
-    loc : Loc.t option; (* debug *)
+| IndexWithDiscriminationTree of {
+    mode : mode;
+    arg_depths : int list;   (* the list of args on which the trie is built *)
+    time : int;         (* time is used to recover the total order *)
+    args_idx : clause DT.t; 
 }
-and mode = bool list (* true=input, false=output *)
 [@@deriving show]
 
 type constraints = stuck_goal list
@@ -167,9 +184,21 @@ type suspended_goal = {
   goal : int * term
 }
 
+(** 
+  Used to index the parameters of a predicate P
+  - [MapOn N] -> N-th argument at depth 1 (head symbol only)
+  - [Hash L]  -> L is the list of depths given by the urer for the parameters of
+                 P. Indexing is done by hashing all the parameters with a non
+                 zero depth and comparing it with the hashing of the parameters
+                 of the query
+  - [DiscriminationTree L] ->
+            we use the same logic of Hash, except we use DiscriminationTree to discriminate
+            clauses
+*)
 type indexing =
   | MapOn of int
   | Hash of int list
+  | DiscriminationTree of int list
 [@@deriving show]
 
 let mkLam x = Lam x [@@inline]
@@ -230,9 +259,14 @@ include Term
 (* Object oriented State.t: borns at compilation time and survives as run time *)
 module State : sig
 
+  type descriptor
+  val new_descriptor : unit -> descriptor
+  val merge_descriptors : descriptor -> descriptor -> descriptor
+
   (* filled in with components *)
   type 'a component
   val declare :
+    descriptor:descriptor ->
     name:string -> pp:(Format.formatter -> 'a -> unit) ->
     init:(unit -> 'a) ->
     clause_compilation_is_over:('a -> 'a) ->
@@ -253,7 +287,8 @@ module State : sig
      - end_compilation (just once before running)
      - end_execution (just once after running)
   *)
-  val init : unit -> t
+  
+  val init : descriptor -> t
   val end_clause_compilation : t -> t
   val begin_goal_compilation : t -> t
   val end_goal_compilation : uvar_body StrMap.t -> t -> t
@@ -267,17 +302,18 @@ module State : sig
   val update_return : 'a component -> t -> ('a -> 'a * 'b) -> t * 'b
   val pp : Format.formatter -> t -> unit
 
+  val dummy : t
+
 end = struct
 
   type stage =
+    | Dummy
     | Compile_prog
     | Compile_goal
     | Link
     | Run
     | Halt
 
-  type t = Obj.t StrMap.t * stage
-
   type 'a component = string
   type extension = {
     init : unit -> Obj.t;
@@ -288,24 +324,38 @@ end = struct
     end_exec : Obj.t -> Obj.t option;
     pp   : Format.formatter -> Obj.t -> unit;
   }
-  let extensions : extension StrMap.t ref = ref StrMap.empty
+  type descriptor = extension StrMap.t ref
 
-  let get name (t,_) =
+  type t = { data : Obj.t StrMap.t; stage : stage; extensions : descriptor }
+  let dummy : t = { data = StrMap.empty; stage = Dummy; extensions = ref StrMap.empty }
+  let descriptor { extensions = x } = x
+ 
+  let new_descriptor () : descriptor = ref StrMap.empty
+  let merge_descriptors m1 m2 =
+    ref (StrMap.merge (fun n e1 e2 ->
+      match e1, e2 with
+      | None, None -> None
+      | Some x, None -> Some x
+      | None, Some x -> Some x
+      | Some _, Some _ -> error ("The state cannot contain two components named "^n) )
+      !m1 !m2)
+
+  let get name { data = t } =
     try Obj.obj (StrMap.find name t)
     with Not_found ->
        anomaly ("State.get: component " ^ name ^ " not found")
 
-  let set name (t,s) v = StrMap.add name (Obj.repr v) t, s
-  let drop name (t,s) = StrMap.remove name t, s
-  let update name (t,s) f =
-    StrMap.add name (Obj.repr (f (Obj.obj (StrMap.find name t)))) t, s
+  let set name ({ data } as x) v = { x with data = StrMap.add name (Obj.repr v) data }
+  let drop name ({ data } as x) = { x with data = StrMap.remove name data }
+  let update name ({ data } as x) f =
+    { x with data = StrMap.add name (Obj.repr (f (Obj.obj (StrMap.find name data)))) data }
   let update_return name t f =
     let x = get name t in
     let x, res = f x in
     let t = set name t x in
     t, res
 
-  let declare ~name ~pp ~init ~clause_compilation_is_over ~goal_compilation_begins ~goal_compilation_is_over ~compilation_is_over ~execution_is_over =
+  let declare ~descriptor:extensions ~name ~pp ~init ~clause_compilation_is_over ~goal_compilation_begins ~goal_compilation_is_over ~compilation_is_over ~execution_is_over =
     if StrMap.mem name !extensions then
       anomaly ("Extension "^name^" already declared");
     extensions := StrMap.add name {
@@ -320,41 +370,61 @@ end = struct
       !extensions;
     name
 
-  let init () =
-    StrMap.fold (fun name { init } acc ->
+  let init extensions : t =
+    let data = StrMap.fold (fun name { init } acc ->
       let o = init () in
       StrMap.add name o acc)
-      !extensions StrMap.empty, Compile_prog
+      !extensions StrMap.empty in
+    {
+      data;
+      stage = Compile_prog;
+      extensions;
+    }
 
-  let end_clause_compilation (m,s) = assert(s = Compile_prog);
-    StrMap.fold (fun name obj acc ->
-      let o = (StrMap.find name !extensions).end_clause obj in
-      StrMap.add name o acc) m StrMap.empty, s
+  let end_clause_compilation { data = m; stage = s; extensions } : t =
+    assert(s = Compile_prog);
+    { data = StrMap.fold (fun name obj acc ->
+        let o = (StrMap.find name !extensions).end_clause obj in
+        StrMap.add name o acc) m StrMap.empty;
+      stage = s;
+      extensions }
 
-  let begin_goal_compilation (m,s) = assert(s = Compile_prog);
-    StrMap.fold (fun name obj acc ->
+  let begin_goal_compilation { data = m; stage = s; extensions } : t =
+    assert(s = Compile_prog);
+    { data = StrMap.fold (fun name obj acc ->
       let o = (StrMap.find name !extensions).begin_goal obj in
-      StrMap.add name o acc) m StrMap.empty, Compile_goal
+      StrMap.add name o acc) m StrMap.empty;
+    stage = Compile_goal;
+    extensions }
 
-  let end_goal_compilation args (m,s) = assert(s = Compile_goal);
-    StrMap.fold (fun name obj acc ->
-      match (StrMap.find name !extensions).end_goal ~args obj with
-      | None -> acc
-      | Some o -> StrMap.add name o acc) m StrMap.empty, Link
+  let end_goal_compilation args { data = m; stage = s; extensions } : t =
+    assert(s = Compile_goal);
+    { data = StrMap.fold (fun name obj acc ->
+        match (StrMap.find name !extensions).end_goal ~args obj with
+        | None -> acc
+        | Some o -> StrMap.add name o acc) m StrMap.empty;
+      stage =  Link;
+      extensions }
 
-  let end_compilation (m,s) = assert(s = Link);
-    StrMap.fold (fun name obj acc ->
-      match (StrMap.find name !extensions).end_comp obj with
-      | None -> acc
-      | Some o -> StrMap.add name o acc) m StrMap.empty, Run
+  let end_compilation { data = m; stage = s; extensions } : t =
+    assert(s = Link);
+    { data = StrMap.fold (fun name obj acc ->
+        match (StrMap.find name !extensions).end_comp obj with
+        | None -> acc
+        | Some o -> StrMap.add name o acc) m StrMap.empty;
+      stage = Run;
+      extensions }
 
-  let end_execution (m,s) = assert(s = Run);
-    StrMap.fold (fun name obj acc ->
-      match (StrMap.find name !extensions).end_exec obj with
-      | None -> acc
-      | Some o -> StrMap.add name o acc) m StrMap.empty, Halt
+  let end_execution { data = m; stage = s; extensions } : t =
+    assert(s = Run);
+    { data = StrMap.fold (fun name obj acc ->
+        match (StrMap.find name !extensions).end_exec obj with
+        | None -> acc
+        | Some o -> StrMap.add name o acc) m StrMap.empty;
+      stage = Halt;
+      extensions }
 
-  let pp fmt (t,s) =
+  let pp fmt { data = t; stage = s; extensions } : unit =
     StrMap.iter (fun name { pp } ->
       try pp fmt (StrMap.find name t)
       with Not_found -> ())
@@ -362,6 +432,8 @@ end = struct
 
 end
 
+let elpi_state_descriptor = State.new_descriptor ()
+
 (* This module contains the symbols reserved by Elpi and the ones
    declared by the API client via declare_global_symbol statically
    (eg the API must be called in a OCaml toplevel value).
@@ -614,8 +686,18 @@ module Conversion = struct
     | RawGoal of term
   type extra_goals = extra_goal list
   type extra_goals_postprocessing = extra_goals -> State.t -> State.t * extra_goals
-  let extra_goals_postprocessing : extra_goals_postprocessing ref = ref (fun x s -> s, x)
   
+  let extra_goals_postprocessing : extra_goals_postprocessing State.component = State.declare
+    ~descriptor:elpi_state_descriptor
+    ~name:"elpi:extra_goals_postprocessing"
+    ~pp:(fun _ _ -> ())
+    ~clause_compilation_is_over:(fun b -> b)
+    ~goal_compilation_begins:(fun b -> b)
+    ~goal_compilation_is_over:(fun ~args:_ b -> Some b)
+    ~compilation_is_over:(fun x -> Some x)
+    ~execution_is_over:(fun x -> Some x)
+    ~init:(fun () -> (); fun x s -> s, x)
+
   type ty_ast = TyName of string | TyApp of string * ty_ast * ty_ast list
   [@@deriving show]
 
@@ -638,11 +720,29 @@ module Conversion = struct
 
   exception TypeErr of ty_ast * int * term (* a type error at data conversion time *)
 
-let rec show_ty_ast ?(outer=true) = function
+type prec_level =
+  | Arrow
+  | AppArg
+
+let need_par x y =
+  match x,y with
+  | Some AppArg, Arrow -> true
+  | Some AppArg, AppArg -> true
+  | Some Arrow, Arrow -> true
+  | Some Arrow, AppArg -> false
+  | None, _ -> false
+
+let with_par p1 p2 s = if need_par p1 p2 then "("^s^")" else s
+
+let rec show_ty_ast ?prec = function
   | TyName s -> s
+  | TyApp ("->",src,[tgt]) ->
+      let src = show_ty_ast ~prec:Arrow src in
+      let tgt = show_ty_ast tgt in
+      with_par prec Arrow (src ^" -> "^ tgt)
   | TyApp (s,x,xs) ->
-      let t = String.concat " " (s :: List.map (show_ty_ast ~outer:false) (x::xs)) in
-      if outer then t else "("^t^")"
+      let t = String.concat " " (s :: List.map (show_ty_ast ~prec:AppArg) (x::xs)) in
+      with_par prec AppArg t
 
 let term_of_extra_goal = function
   | Unify(a,b) -> Builtin(Global_symbols.eqc,[a;b])
@@ -738,7 +838,9 @@ module ContextualConversion = struct
 
   end
 
-let while_compiling = State.declare ~name:"elpi:compiling"
+let while_compiling : bool State.component = State.declare
+  ~descriptor:elpi_state_descriptor
+  ~name:"elpi:compiling"
   ~pp:(fun fmt _ -> ())
   ~clause_compilation_is_over:(fun b -> b)
   ~goal_compilation_begins:(fun b -> b)
@@ -747,6 +849,83 @@ let while_compiling = State.declare ~name:"elpi:compiling"
   ~execution_is_over:(fun _ -> Some false) (* we keep it, since API.FlexibleData.Elpi.make needs it *)
   ~init:(fun () -> false)
 
+module HoasHooks = struct
+
+type descriptor = {
+  extra_goals_postprocessing: Conversion.extra_goals_postprocessing option;
+}
+
+let new_descriptor () = ref {
+  extra_goals_postprocessing = None;
+}
+
+let set_extra_goals_postprocessing ~descriptor f =
+  match !descriptor with
+  | { extra_goals_postprocessing = None } ->
+     descriptor := { extra_goals_postprocessing = Some f }
+  | { extra_goals_postprocessing = Some _ } ->
+      error "set_extra_goals_postprocessing called twice"
+
+end
+
+module CalcHooks = struct
+
+type run = term list -> term
+type eval = { code : run; ty_decl : string; }
+type descriptor = (constant * eval) list
+let new_descriptor () : descriptor ref = ref []
+
+let eval : run Constants.Map.t State.component =
+  State.declare ~descriptor:elpi_state_descriptor ~name:"elpi:eval"
+  ~clause_compilation_is_over:(fun x -> x)
+  ~goal_compilation_begins:(fun x -> x)
+  ~goal_compilation_is_over:(fun ~args:_ x -> Some x)
+  ~compilation_is_over:(fun x -> Some x)
+  ~execution_is_over:(fun _ -> None)
+  ~init:(fun () -> Constants.Map.empty)
+  ~pp:(fun fmt t -> Constants.Map.pp (fun _ _ -> ()) fmt t)
+
+end
+
+module QuotationHooks = struct
+  
+type quotation = depth:int -> State.t -> Loc.t -> string -> State.t * term
+
+type descriptor = {
+  named_quotations : quotation StrMap.t;
+  default_quotation : quotation option;
+  singlequote_compilation : (string * (State.t -> F.t -> State.t * term)) option;
+  backtick_compilation : (string * (State.t -> F.t -> State.t * term)) option;
+}
+
+let new_descriptor () = ref {
+  named_quotations = StrMap.empty;
+  default_quotation = None;
+  singlequote_compilation = None;
+  backtick_compilation = None;
+}
+
+let declare_singlequote_compilation ~descriptor name f =
+  match !descriptor with
+  | { singlequote_compilation = None } ->
+      descriptor := { !descriptor with singlequote_compilation = Some(name,f) }
+  | { singlequote_compilation = Some(oldname,_) } ->
+        error("Only one custom compilation of 'ident' is supported. Current: "
+          ^ oldname ^ ", new: " ^ name)
+let declare_backtick_compilation ~descriptor name f =
+  match !descriptor with
+  | { backtick_compilation = None } ->
+      descriptor := { !descriptor with backtick_compilation = Some(name,f) }
+  | { backtick_compilation = Some(oldname,_) } ->
+        error("Only one custom compilation of `ident` is supported. Current: "
+          ^ oldname ^ ", new: " ^ name)
+
+let set_default_quotation ~descriptor f =
+  descriptor := { !descriptor with default_quotation = Some f }
+let register_named_quotation ~descriptor ~name:n f =
+  descriptor := { !descriptor with named_quotations = StrMap.add n f !descriptor.named_quotations }  
+
+end
 module BuiltInPredicate = struct
 
 type name = string
@@ -767,6 +946,7 @@ type ('function_type, 'inernal_outtype_in, 'internal_hyps, 'internal_constraints
   | Easy : doc -> (depth:int -> 'o, 'o,unit,unit) ffi
   | Read : ('h,'c) ContextualConversion.ctx_readback * doc -> (depth:int -> 'h -> 'c -> State.t -> 'o, 'o,'h,'c) ffi
   | Full : ('h,'c) ContextualConversion.ctx_readback * doc -> (depth:int -> 'h -> 'c -> State.t -> State.t * 'o * Conversion.extra_goals, 'o,'h,'c) ffi
+  | FullHO : ('h,'c) ContextualConversion.ctx_readback * doc -> (once:(depth:int -> term -> State.t -> State.t) -> depth:int -> 'h -> 'c -> State.t -> State.t * 'o * Conversion.extra_goals, 'o,'h,'c) ffi
   | VariadicIn    : ('h,'c) ContextualConversion.ctx_readback * ('t,'h,'c) ContextualConversion.t * doc -> ('t list -> depth:int -> 'h -> 'c -> State.t -> State.t * 'o, 'o,'h,'c) ffi
   | VariadicOut   : ('h,'c) ContextualConversion.ctx_readback * ('t,'h,'c) ContextualConversion.t * doc -> ('t oarg list -> depth:int -> 'h -> 'c -> State.t -> State.t * ('o * 't option list option), 'o,'h,'c) ffi
   | VariadicInOut : ('h,'c) ContextualConversion.ctx_readback * ('t ioarg,'h,'c) ContextualConversion.t * doc -> ('t ioarg list -> depth:int -> 'h -> 'c -> State.t -> State.t * ('o * 't option list option), 'o,'h,'c) ffi
@@ -1079,6 +1259,7 @@ let pp_tab_args fmt l =
   let n = List.length l - 1 in
   let max = List.fold_left (fun m (_,s,_) -> max (String.length s) m) 0 l in
   Fmt.pp_open_tbox fmt ();
+  if l = [] then Fmt.fprintf fmt ".";
   List.iteri (fun i x ->
     let sep = if i = n then "." else "," in
     pp_tab_arg i max sep fmt x) l;
@@ -1129,6 +1310,7 @@ let document_pred fmt docspec name ffi =
     | Read (_,s) -> pp_pred fmt docspec name s args
     | Easy s -> pp_pred fmt docspec name s args
     | Full (_,s) -> pp_pred fmt docspec name s args
+    | FullHO (_,s) -> pp_pred fmt docspec name s args
     | VariadicIn( _,{ ContextualConversion.ty }, s) -> pp_variadictype fmt name s (Conversion.show_ty_ast ty) args
     | VariadicOut( _,{ ContextualConversion.ty }, s) -> pp_variadictype fmt name s (Conversion.show_ty_ast ty) args
     | VariadicInOut( _,{ ContextualConversion.ty }, s) -> pp_variadictype fmt name s (Conversion.show_ty_ast ty) args
@@ -1136,13 +1318,18 @@ let document_pred fmt docspec name ffi =
     doc [] ffi
 ;;
 
-let document fmt l =
+let document fmt l calc_list =
   let omargin = Fmt.pp_get_margin fmt () in
   Fmt.pp_set_margin fmt 75;
   Fmt.fprintf fmt "@[<v>";
   Fmt.fprintf fmt "@\n@\n";
   List.iter (function
-    | MLCode(Pred(name,ffi,_), docspec) -> document_pred fmt docspec name ffi
+    | MLCode(Pred(name,ffi,_), docspec) ->
+        document_pred fmt docspec name ffi;
+        if name = "calc" then begin
+          Format.fprintf fmt "%s@\n@\n" "%  --- Operators ---";
+          List.iter (fun (_,x) -> Format.fprintf fmt "%s@\n@\n" x.CalcHooks.ty_decl ) calc_list
+        end;
     | MLData { pp_doc } -> Fmt.fprintf fmt "%a@\n" pp_doc ()
     | MLDataC { pp_doc } -> Fmt.fprintf fmt "%a@\n" pp_doc ()
     | LPCode s -> Fmt.fprintf fmt "%s" s; Fmt.fprintf fmt "@\n@\n"
@@ -1207,6 +1394,7 @@ type 'a solution = {
   state : State.t;
   output : 'a;
   pp_ctx : pp_ctx;
+  state_for_relocation : int * State.t;
 }
 type 'a outcome = Success of 'a solution | Failure | NoMoreSteps
 
