@@ -72,7 +72,7 @@ type raw_attribute =
   | Before of string
   | Replace of string
   | External
-  | Index of int list
+  | Index of int list * string option
 [@@ deriving show]
 
 module Clause : sig
@@ -213,7 +213,8 @@ and cattribute = {
 }
 and tattribute =
   | External
-  | Index of int list
+  | Index of int list * tindex option
+and tindex = Map | HashMap | DiscriminationTree
 and 'a shorthand = {
   iloc : Loc.t;
   full_name : 'a;
