@@ -437,4 +437,29 @@ let () = declare "trace-browser-elab-broken2"
   ~description:"fatal broken trace elaboration"
   ~expectation:(FailureOutput (Str.regexp "broken.*step_id 217.*json object 1857"))
   ()
+  
+
+let () = declare "bad_index"
+  ~source_elpi:"bad_index.elpi"
+  ~description:"bad indexing directive"
+  ~expectation:Test.(FailureOutput (Str.regexp "Wrong indexing"))
+  ()
+ 
+
+let sample = mk_tmp_file "trace_w.json" ".new"
+let () = declare "trace-browser-w"
+  ~source_elpi:"trace-w/main.elpi"
+  ~description:"trace generation"
+  ~typecheck:false
+  ~trace:(On["json";"file://"^sample;"-trace-at";"0";"99";"-trace-only";"user"])
+  ~expectation:(SuccessOutputFile { sample; adjust = Util.strip_cwd; reference = "trace_w.json" })
+  ()
+
+let sample = mk_tmp_file "trace_w.elab.json" ".new"
+let () = declare "trace-browser-w-elab"
+  ~source_json:"trace_w.json"
+  ~description:"trace elaboration"
+  ~expectation:(SuccessOutputFile { sample; adjust = Util.strip_cwd; reference = "trace_w.elab.json" })
+  ()
+  
   
\ No newline at end of file
